#MIT License
#
#Copyright (c) [2024] Lukas Walker
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:
#
#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

'''
``General information:``\n
This package is meant to program small webapps real quick without much required knowledge
docmentation can be found on <https://bbwebservice.eu>
'''

import os
import sys


default_config = '''
{
    "ip": "default",
    "port": 5000,
    "queue_size": 10,
    "max_threads": 100,
	"SSL": false,
 	"host" : "",
	"cert_path" : "",
	"key_path" : ""
}
'''


MAIN_PATH = os.path.dirname(os.path.abspath(sys.argv[0]))
if not os.path.exists(MAIN_PATH + "/content"):
    os.makedirs(MAIN_PATH + "/content")
if not os.path.exists(MAIN_PATH + "/config"):
    os.makedirs(MAIN_PATH + "/config")
if not os.path.exists(MAIN_PATH + "/config/config.json"):
    with open(MAIN_PATH + "/config/config.json", 'w') as config:
        config.writelines(str(default_config))



#2.0

import os
import re
from . import core
from .__init__ import MAIN_PATH
from .app_utils import urlencoded_to_dict
from .http_parser import Redirect
from .url_utils import UrlTemplate 
from .http_parser import PartialContent
from .http_parser import HTTP_Status_Code
from .http_parser import HTTP_Response as Response
from .http_parser import HTTP_Message_Header_Line as Header_Line
from .http_parser import HTTP_Message_Request_Header_Tag as Request_Header_Tag
from .http_parser import HTTP_Message_Response_Header_Tag as Response_Header_Tag

class MIME_TYPE:
    HTML = 'text/html'
    JAVA_SCRIPT = 'text/javascript'
    CSS = 'text/css'
    ICO = "image/x-icon"
    PNG = "image/png"
    SVG = "image/svg+xml"
    TEXT = "text/plain"
    MP4 = "video/mp4"
    JSON = "application/json"
    WEBM_AUDIO = "audio/webm"
    ICO = "image/x-icon"

class STORE_VARS:
    #TODO: UPDATE Vars
    '''Keys used to access attributes of the dict provided to the respective handler methods'''

    COOKIES = 'cookies'
    QUERY_STRING = 'query_string'
    RESPONSE = 'response'
    POST = 'post'
    FLAGS = 'flags'
    REQUEST_HEADER = 'request_header'
    TEMPLATE_VARS = 'template_args'
    

class LOGGING_OPTIONS:

    '''Options used to set the logging mode via the ``set_logging`` function'''

    REQUESTS = 'request'
    RESPONSES = 'response'
    DEBUG = 'debug'
    INFO = 'info'
    WARNING = 'warning'
    ERROR = 'error'
    CRITICAL = 'critical'
    TIME = 'time'

def register(*args, **kwargs):

    '''
    ``General Information:``\n
    The register decorator adds the decorated function to the servers PAGES
    dictionary. The decorater requires a path and a type argument to be specified. The decorated function
    gets called whenever a ``GET request`` is targeted to the speciefied path. The function is expected to return a String
    that's been served under the specified route the type argument must provide the matching MIME-type. If the decorated function
    takes an argument it gets passed a session specific dict.

    ``Usage:``
    ```python
    @register(route= '/', type= MIME_TYPE.TEXT)
    def main(args):
        return str(args)
    ```
    ``Note:`` The ``args`` argument is optional
    '''

    if 'route' not in kwargs:
        raise Exception('The "route" argument is missing.')
    if 'type' not in kwargs:
        raise Exception('The "type" argument is missing.')
    route = kwargs['route']
    def inner(func):
        if isinstance(route, str) and route not in core.PAGES:
                core.PAGES[route] = [func,kwargs['type']]
        
        if isinstance(route, UrlTemplate) and route not in core.GET_TEMPLATES:
                route.type = kwargs['type']
                route.handler = func
                core.GET_TEMPLATES.append(route)
        return func            
    return inner

def post_handler(*args, **kwargs):

    '''
    ``General Information:``\n
    The post_handler decorator adds the decorated function to the servers POST_HANDLER
    dictionary. The decorater requires a path and a type argument to be specified the decorated function
    gets called whenever a post is targeted to the speciefied path. The function is expected to return a string of the
    under the type attribute declared MIME-type. The function decorated gets passed a dictionary containing
    the variables passed with the ``POST so it has to take an argument``
    
    ``Usage:``
    ```python
    @post_handler(route= '/', type= MIME_TYPE.TEXT)
    def main(args):
        return str(args)
    ```
    ``Note:`` The ``args`` argument is necessary!
    '''
    

    if 'route' not in kwargs:
        raise Exception('The "route" argument is missing.')
    if 'type' not in kwargs:
        raise Exception('The "type" argument is missing.')
    route = kwargs['route']
    def inner(func):
        if func.__code__.co_argcount != 1:
            raise Exception('The function decorated with the post_handler decorater has to\n accept one Argument "args":dict')

        if isinstance(route, str) and route not in core.POST_HANDLER:
                core.POST_HANDLER[route] = [func,kwargs['type']]
                
        if isinstance(route, UrlTemplate) and route not in core.POST_TEMPLATES:
                route.type = kwargs['type']
                route.handler = func
                core.POST_TEMPLATES.append(route)
        return func                  
    return inner


def error_handler(*args, **kwargs):

    '''
    ``General Information:``\n
    The error_handler decorator is used to serve a specific page if an error-code occurs.
    ``
    
    ``Usage:``
    ```python
    @error_handler(error_code= 404, type= MIME_TYPE.TEXT)
    def main():
        return str('Page not found!')
    ```
    '''
    

    if 'error_code' not in kwargs:
        raise Exception('The "error_code" argument is missing.')
    if 'type' not in kwargs:
        raise Exception('The "type" argument is missing.')
    err = kwargs['error_code']
    def inner(func):
        if func.__code__.co_argcount != 0:
            raise Exception('The function decorated with the error_handler is not allowed to take arguments.')
        if kwargs['error_code'] not in core.ERROR_HANDLER:
                core.ERROR_HANDLER[err] = [func,kwargs['type']]
        return func            
    return inner

def load_file(path:str) -> str:

    '''The load_file function attempts to read the content of a file with given path and returns it as a utf-8 encoded string'''

    with open(MAIN_PATH+path,'r',encoding='utf-8') as content:
            lines = content.readlines()
            return ''.join(lines)

def load_bin_file(path:str) -> bytes:

    '''The load_file function attempts to read the content of a file with given path and returns it as bytes'''

    path = MAIN_PATH+path
    size = os.path.getsize(path)
    with open(path,'rb') as content:
            return content.read(size)

#TODO: File through new file not found error so status code 404 can be returned
def load_file_from_directory(root_path: str, file_path: str) -> str:
    """The load_file_from_directory function attempts to read the content of a file with given path 
    relative to root_path and returns it as a utf-8 encoded string.
    """
    
    abs_root_path = os.path.abspath(root_path)
    abs_file_path = os.path.abspath(os.path.join(root_path, file_path))
    if os.path.commonpath([abs_root_path]) != os.path.commonpath([abs_root_path, abs_file_path]):
        raise ValueError("Attempt to access file outside the root directory")
    
    with open(abs_file_path, 'r', encoding='utf-8') as content:
        return content.read()

def load_bin_file_from_directory(root_path: str, file_path: str) -> bytes:
    """The load_bin_file_from_directory function attempts to read the content of a file with given path 
    relative to root_path and returns it as bytes.
    """
    
    abs_root_path = os.path.abspath(root_path)
    abs_file_path = os.path.abspath(os.path.join(root_path, file_path))
    

    if os.path.commonpath([abs_root_path]) != os.path.commonpath([abs_root_path, abs_file_path]):
        raise ValueError("Attempt to access file outside the root directory")
    
    with open(abs_file_path, 'rb') as content:
        return content.read()



def render_page(path:str, args:dict) -> str:

    '''The render_page function reads a files content and executes python-code wrapped in double curly braces "{{".
    If a variable name starts with an underscore "_" the python code wrapped in curley braces will be substituded by it's content.
    The values passed with the args dictionary can be accessed with the globals() function in the targeted file'''

    content = ''
    try:
        with open(MAIN_PATH+path,'r',encoding='utf-8') as page:
            content = '\n'.join(page.readlines())
            res_content = content[:]
        to_eval = re.finditer("{{(.|\n)*?}}",content)
        for var in to_eval:
            OUTPUT = {}
            var = content[var.span()[0]:var.span()[1]]
            exec(compile(var.strip('{{}}'),'temp','exec'),args,OUTPUT)
            insert = '\n'.join([str(OUTPUT[x]) for x in OUTPUT.keys() if x[0] == '_'])
            res_content = res_content.replace(var,insert)
    except Exception as e:
        print('[RENDERER] Error with rendering Page.')
        print(e)
    return res_content

def substitude_vars(content:str, vars:dict) -> str:
    
    '''This function accepts a string and substitudes all "%% + key + %%" with the according value from the dictionary'''
    
    try:
        _content = content
        for key in vars:
            _content = re.sub(f'%%{key}%%',vars[key],_content)
    except:
        print('[SUBSTITUDE_VARS] Error substituding vars.')
    return _content

def set_cookie(args, key, value) -> bool:
    if key not in args[STORE_VARS.COOKIES]:
        res:Response = args[STORE_VARS.RESPONSE]
        res.header.add_header_line(Header_Line(Response_Header_Tag.SET_COOKIE,f'{key}={value}'))
        return True
    return False


def is_partial(args) -> bool:
    return 'partial' in args[STORE_VARS.FLAGS]


def set_logging(option:str, state:bool) -> None:
    if option in core.LOGGING_OPTIONS:
        core.LOGGING_OPTIONS[option] = state
        print(f'[LOGGING] logging-option "{option}" set to {state}.')
    else:
        print(f'[LOGGING] Error: There is no logging-option called "{option}".')

def set_logging_callback(callback):
    try:
        assert callable(callback), 'The logging callback is not a callable'
        assert callback.__code__.co_argcount == 3, 'The logging callback needs to accept 3 arguments (message, time_stamp, loglvl)'
        core.LOGGING_CALLBACK.append(callback)
        print('[LOGGING]', "Logging callback has been set")
        
    except Exception as e:
        print('[LOGGING]','Error:',e)

def log_to_file(path='/log.txt', logging_options=[LOGGING_OPTIONS.DEBUG]):
    try:
        for option in logging_options:
            assert option in core.LOGGING_OPTIONS, f'log_to_file: There is no logging-option called "{option}".'

        def file_writer_callback(message, time_stamp, loglvl):
            if  loglvl in logging_options:
                with open(MAIN_PATH+path,'a') as file:
                    file.write(f'({time_stamp}) {message}\n')
        core.LOGGING_CALLBACK.append(file_writer_callback)
        print('[LOGGING]', f"File-logging has been activated path= {path}")
        
    except Exception as e:
        print('[LOGGING]', e)
        
    
 
def get_pages() -> dict:
    return core.PAGES

def get_sessions() -> dict:
    return core.SESSIONS
    
def get_post_handler() -> dict:
    return core.POST_HANDLER

def start() -> None:

    '''The start function causes the server to start by invoking core.start()'''

    core.start()



import os 
from . import MAIN_PATH


def load_bin_file_partial(path, start_byte, end_byte) -> bytes:

    path = MAIN_PATH+path
    with open(path, 'rb') as file:
        file.seek(start_byte)
        data = file.read(end_byte - start_byte)
    return data


def get_file_size(path: str) -> int:
    try:
        size = os.path.getsize(MAIN_PATH+path)
        return size
    except FileNotFoundError:
        return -1


class PartialContent:

    def __init__(self, path, default_size) -> None:
        self.path = path
        self.default_size = default_size
    
    def get_range(self, start, end) -> bytes:
        
        end = end if end else end + self.default_size
        return load_bin_file_partial(self.path, start,min(self.get_size(), end+1))
    
    def get_size(self) -> int:
        return get_file_size(self.path)
    

class Redirect:
    
    def __init__(self, path, status= None) -> None:
        self.path = path
        self.status = status 



import re
HEX_CHARS = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F']


#TODO: unefficient af fix that 

def leading_ones_count(encoded_byte):
    byte_value = int(encoded_byte, 16)
    count = 0
    mask = 0x80 
    while mask != 0 and byte_value & mask:
        count += 1
        mask >>= 1
    return count

def unescape_url(url:str) -> str:
    unescaped = ''
    escape_sequence = ''
    index = 0
    count = 0
    url_len = len(url)
    
    while index < url_len:
        
        if len(escape_sequence) == 0 and url[index] == '%' and index + 3 <= url_len and url[index+1] in HEX_CHARS and url[index+2] in HEX_CHARS:
            escape_byte = url[index+1:index+3]
            count =  leading_ones_count(escape_byte)
            escape_sequence += escape_byte
            index += 3
            
        elif url[index] == '%' and index + 3 <= url_len and url[index+1] in HEX_CHARS and url[index+2] in HEX_CHARS and count > 1:
            escape_byte = url[index+1:index+3]
            count -= 1
            escape_sequence+= escape_byte
            index += 3
            
        elif escape_sequence:
            unescaped += decode_hex_string(escape_sequence)
            escape_sequence = ''
            count = 0
        else:
            unescaped += url[index]
            index += 1

    return unescaped if not escape_sequence else unescaped + decode_hex_string(escape_sequence)
            

def decode_hex_string(hex_string:str) ->str:
    try:
        decoded_bytes = bytes.fromhex(hex_string)
        unicode_char = decoded_bytes.decode('utf-8')
        return unicode_char
    except UnicodeDecodeError:
        return ''


#TODO: native URL matching withoug re for better performance 
#TODO: special hashing where string gets hash of matching template 
import re

class UrlTemplate:
    def __init__(self, template_string):
        self.template = template_string
        self.regex_pattern = '^' + re.sub(r'\{(\w+):(\w+)\}', self._repl, template_string) + '$'
        self.handler = None
        self.type = None
    
    def _repl(self, match):
        type_ = match.group(2)
        if type_ == 'int':
            return r'(\d+)'
        elif type_ == 'str':
            return r'(\w+)'
        elif type_ == 'float':
            return r'(\d+\.\d+)'
        elif type_ == 'bool':
            return r'(True|False)'
        elif type_ == 'path':
            return r'(.+)'
        else:
            raise ValueError(f"Unknown type: {type_}")
        
    def convert(self, value, type_):
        if type_ == 'int':
            return int(value)
        elif type_ == 'float':
            return float(value)
        elif type_ == 'bool':
            return value == 'True'
        else:
            return value
        
    def extract(self, url):
        match = re.match(self.regex_pattern, url)
        if match:
            return {k: self.convert(v, t) for (k, t), v in zip(re.findall(r'\{(\w+):(\w+)\}', self.template), match.groups())}
        else:
            return None
        
    def __eq__(self, url):
        if isinstance(url, str):
            return re.match(self.regex_pattern, url) is not None
        if isinstance(url, self.__class__):
            return self.template == url.template
        
        return False



import socket
from threading import Thread, Event, current_thread
import ssl
import os
from .config_loader import Config
from .http_parser import HTTP_Message_Factory, log, LOGGING_OPTIONS

# Define globals
SESSIONS = {}
PAGES = {}
GET_TEMPLATES = []
POST_HANDLER = {}
POST_TEMPLATES = []
ERROR_HANDLER = {}
SERVER_THREADS = []
CONFIG = Config()

def servlet(conn, addr, worker_state):
    """Handle a client connection in a separate thread."""
    try:
        while worker_state.is_set():
            log(f'[THREADING] thread {current_thread().ident} listens now.', log_lvl='debug')
            
            message_factory = HTTP_Message_Factory(conn, addr, PAGES, GET_TEMPLATES, POST_HANDLER, POST_TEMPLATES, ERROR_HANDLER)
            resp = message_factory.get_message()
            conn.sendall(resp)
            
            header, _, content = resp.partition(b'\r\n\r\n')
            log('\n\nRESPONSE:', str(header, 'utf-8'), content, '\n\n', log_lvl='response', sep='\n')
            
            if not message_factory.stay_alive:
                log(f'[THREADING] thread {current_thread().ident} closes because stay_alive is set to False', log_lvl='debug')
                break

    except TimeoutError:
        log(f'[THREADING] thread {current_thread().ident} closes due to a timeout error.', log_lvl='debug')
    except Exception as err:
        log(f'[THREADING] thread {current_thread().ident} closes due to an error: "{err}"', log_lvl='debug')
    finally:
        try:
            conn.settimeout(1.0)
            conn.close()
        except Exception as e:
            log(f'[THREADING] thread {current_thread().ident} encountered an error while closing connection: {e}', log_lvl='debug')

def main(server, state):
    """Accept and dispatch client connections to servlets."""
    print(f'[SERVER] {CONFIG.SERVER_IP} running on port {CONFIG.SERVER_PORT}...')
    while state.is_set():
        global SERVER_THREADS
        SERVER_THREADS = [t for t in SERVER_THREADS if t[0].is_alive()]

        if len(SERVER_THREADS) >= CONFIG.MAX_THREADS:
            continue

        try:
            conn, addr = server.accept()
            worker_state = Event()
            worker_state.set()
            if conn:
                worker_thread = Thread(target=servlet, args=(conn, addr, worker_state))
                SERVER_THREADS.append([worker_thread, worker_state, conn])
                worker_thread.start()
                conn.settimeout(15)
        except TimeoutError:
            pass
        except Exception as e:
            if state.is_set():
                log(f'[CONNECTION_ERROR] A connection failed due to the following error: {e}.\n', log_lvl='debug')

def shutdown_server(server, server_thread, server_state):
    """Shutdown the server and clean up resources."""
    server_state.clear()
    for worker_thread, worker_state, conn in SERVER_THREADS:
        worker_state.clear()
        try:
            conn.shutdown(socket.SHUT_RDWR)
            conn.close()
        except Exception as e:
            print(f"[SERVER] Error while closing client connection: {e}")
        worker_thread.join(timeout=1)

    try:
        server.shutdown(socket.SHUT_RDWR)
        server.close()
    except Exception as e:
        print(f"[SERVER] Error while closing server: {e}")
    server_thread.join(timeout=1)
    print('[SERVER] Closed...')

#in the future bbwebservice will be wrapped in Klasses so multible server instanzes can be hosted in one process
def sni_callback(sock, server_name, context):
    try:
        if server_name == CONFIG.HOST:
            context.load_cert_chain(certfile=CONFIG.CERT_PATH, keyfile=CONFIG.KEY_PATH)
            log(f'[SNI CALLBACK] Successfully loaded certificate for {server_name}', log_lvl='debug')
        else:
            log(f'[SNI CALLBACK] Unknown server name: {server_name}', log_lvl='debug')
    except Exception as e:
        log(f'[SNI CALLBACK] Error loading certificate: {e}', log_lvl='debug')

def start():
    """Start the server and initiate the main dispatcher."""
    if CONFIG.SERVER_IP == 'default':
        CONFIG.SERVER_IP = socket.gethostbyname(socket.gethostname())
    try:
        socket.setdefaulttimeout(2)
        server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server.bind((CONFIG.SERVER_IP, CONFIG.SERVER_PORT))
        server.listen(CONFIG.QUE_SIZE)
        server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        if CONFIG.SSL:
            SERVER_HOSTNAME = CONFIG.HOST
            context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
            context.load_cert_chain(CONFIG.CERT_PATH, CONFIG.KEY_PATH)
            server = context.wrap_socket(server, server_side=True)
            log(f'[SERVER] ssl active', log_lvl='debug')
            if SERVER_HOSTNAME:
                context.sni_callback = sni_callback
                log(f'[SERVER] SNI is active for {SERVER_HOSTNAME}\n', log_lvl='debug')
    except Exception as e:
        log(f'[SERVER] Error while attempting to start the server: {e}\n', log_lvl='debug')
        os._exit(1)

    server_state = Event()
    server_state.set()
    server_thread = Thread(target=main, args=(server, server_state))
    server_thread.start()

    try:
        while True:
            state = input()
            if state in ['quit', 'q', 'exit', 'e', 'stop']:
                shutdown_server(server, server_thread, server_state)
                break
    except KeyboardInterrupt:
        shutdown_server(server, server_thread, server_state)
        os._exit(0)
